from datetime import datetime
from dataclasses import dataclass, field
from io import StringIO
from typing import List
from collections import namedtuple
from .constants import tick_fields , bar_fields , bid_ask_fields , greek_fields , full_feed_fields, min_fields
from .exceptions import TDLiveCalcError
from colorama import Style, Fore    
from datetime import datetime as dt
import shutil
import pickle
import requests
import os
import pandas as pd

@dataclass
class TouchlineData:
    raw_data : List[any] = field(init = True , repr = False)

    def __post_init__(self):
        self.symbol = self.raw_data[0]
        self.symbol_id = self.raw_data[1]
        self.timestamp = datetime.strptime(self.raw_data[2], '%Y-%m-%dT%H:%M:%S')
        self.ltp = float(self.raw_data[3])
        self.ltq = float(self.raw_data[4])
        self.atp = float(self.raw_data[5])
        self.ttq = int(self.raw_data[6])
        self.day_open = float(self.raw_data[7])
        self.day_high = float(self.raw_data[8])
        self.day_low = float(self.raw_data[9])
        self.prev_day_close = float(self.raw_data[10])
        self.oi = int(self.raw_data[11])
        self.prev_day_oi = int(self.raw_data[12])
        self.turnover = float(self.raw_data[13])
        self.best_bid_price = float(self.raw_data[14])
        self.best_bid_qty = float(self.raw_data[15])
        self.best_ask_price = float(self.raw_data[16])
        self.best_ask_qty = float(self.raw_data[17])
        


@dataclass
class tick_feed:
    touchline: TouchlineData = field(init = True , repr = False)
    
    def __post_init__(self):
        self.timestamp = self.touchline.timestamp
        self.symbol = self.touchline.symbol
        self.symbol_id = self.touchline.symbol_id
        self.ltp = self.touchline.ltp
        self.ltq = self.touchline.ltq
        self.atp = self.touchline.atp
        self.ttq = self.touchline.ttq
        self.day_open = self.touchline.day_open
        self.day_high = self.touchline.day_high
        self.day_low = self.touchline.day_low
        self.prev_day_close = self.touchline.prev_day_close
        self.oi = self.touchline.oi
        self.prev_day_oi = self.touchline.prev_day_oi
        self.turnover = self.touchline.turnover
        self.special_tag = ""
        self.tick_seq = None
        self.best_bid_price = self.touchline.best_bid_price
        self.best_bid_qty = self.touchline.best_bid_qty
        self.best_ask_price = self.touchline.best_ask_price
        self.best_ask_qty = self.touchline.best_ask_qty

    def __repr__(self):
        tick_tuple = namedtuple('tick_feed', field_names= tick_fields)
        attributes = list(map(lambda x: getattr(self , x), tick_fields))
        return str(tick_tuple(*attributes))    

    @property
    def change(self):
        try:
            return self.ltp - self.prev_day_close
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} and tick_seq={self.tick_seq}")
            
    @property
    def oi_change(self):
        try:
            return self.oi - self.prev_day_oi
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} and tick_seq={self.tick_seq}")
       
    @property
    def change_perc(self):
        try:
            return self.change * 100 / self.prev_day_close
        except ZeroDivisionError:
            return 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} and tick_seq={self.tick_seq}")
   
    @property
    def oi_change_perc(self):
        try:
            return self.oi_change * 100 / self.prev_day_oi
        except ZeroDivisionError:
            return 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} and tick_seq={self.tick_seq}")
        
@dataclass
class min_feed:

    touchline: TouchlineData = field(init = True , repr = False)
    
    def __post_init__(self):
        self.symbol = self.touchline.symbol
        self.symbol_id = self.touchline.symbol_id
        self.day_open = self.touchline.day_open
        self.day_high = self.touchline.day_high
        self.day_low = self.touchline.day_low
        self.prev_day_close = self.touchline.prev_day_close
        self.prev_day_oi = self.touchline.prev_day_oi
        self.oi = self.touchline.oi
        self.ttq = self.touchline.ttq
        self.timestamp = None
        self.open = None
        self.high = None
        self.low = None
        self.close = None
        self.volume = None

    def __repr__(self):
        min_feed_tuple = namedtuple('min_feed', field_names= min_fields)
        attributes = list(map(lambda x: getattr(self , x), min_fields))
        return str(min_feed_tuple(*attributes))    
    
    @property
    def change(self):
        try:
            return self.close - self.prev_day_close
        except TypeError:
            return 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with timestamp={self.timestamp}")
       
    @property
    def oi_change(self):
        try:
            return self.oi - self.prev_day_oi
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with timestamp={self.timestamp}")

    @property
    def change_perc(self):
        try:
            return self.change * 100 / self.prev_day_close
        except ZeroDivisionError:
            return 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with timestamp={self.timestamp}")

    @property
    def oi_change_perc(self):
        try:
            return self.oi_change * 100 / self.prev_day_oi
        except ZeroDivisionError:
            return 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with timestamp={self.timestamp}")


@dataclass
class full_feed:
    raw_tick : List[any] = field(init = True , repr = False )
    symbol : str = field(init = True )

    def __post_init__(self):
        self.symbol_id = int(self.raw_tick[0])
        self.timestamp = datetime.strptime(self.raw_tick[1], '%Y-%m-%dT%H:%M:%S')
        self.ltp = float(self.raw_tick[2])
        self.ltq = int (self.raw_tick[3])
        self.atp = float(self.raw_tick[4])
        self.ttq = float(self.raw_tick[5])
        self.day_open =  float(self.raw_tick[6])
        self.day_high = float(self.raw_tick[7])
        self.day_low = float(self.raw_tick[8])
        self.prev_day_close = float(self.raw_tick[9])
        self.oi = int(self.raw_tick[10])
        self.prev_day_oi = int(self.raw_tick[11])
        self.turnover = float(self.raw_tick[12])
        self.tick_seq = int(self.raw_tick[14])
        if len(self.raw_tick) > 15:
            self.best_bid_price = float(self.raw_tick[15])
            self.best_bid_qty   =  int(float(self.raw_tick[16]))
            self.best_ask_price = float(self.raw_tick[17])
            self.best_ask_qty   = int(float(self.raw_tick[18]))
        else:
            self.best_bid_price = 0
            self.best_bid_qty   =  0
            self.best_ask_price = 0
            self.best_ask_qty   = 0

    def __repr__(self):
        full_feed_tuple = namedtuple('full_feed', field_names= full_feed_fields)
        attributes = list(map(lambda x: getattr(self , x), full_feed_fields))
        return str(full_feed_tuple(*attributes))    
    

@dataclass
class bidask_feed:
    raw_tick : List[any] = field(init = True , repr = False )
    level : str = field(init = True , repr = False )
    symbol : str = field(init = True )

    def __post_init__(self):
        self.symbol_id = int(self.raw_tick[0])
        self.timestamp = datetime.strptime(self.raw_tick[1], '%Y-%m-%dT%H:%M:%S')
        if self.level == "L1":
            self.bid = [ ( float(self.raw_tick[2]) , int(float(self.raw_tick[3])) ) ]
            self.ask = [ ( float(self.raw_tick[4]) , int(float(self.raw_tick[5])) ) ]
            self.total_bid = None
            self.total_ask = None
        elif self.level == "L2":
            bidask = list(zip(*[iter(self.raw_tick[3:])]*3) )
            self.bid = bidask[0:5]
            self.ask = bidask[5:]
            self.total_bid = self.raw_tick[-2]
            self.total_ask = self.raw_tick[-1]
    
    def __repr__(self):
        bid_ask_tuple = namedtuple('bidask_feed', field_names= bid_ask_fields)
        attributes = list(map(lambda x: getattr(self , x), bid_ask_fields))
        return str(bid_ask_tuple(*attributes))    


@dataclass
class bar_feed:
    raw_tick : List[any]= field(init = True , repr = False )
    symbol :str = field(init = True)
    
    def __post_init__(self):
        self.symbol_id  = int(self.raw_tick[1])
        self.timestamp  = datetime.strptime(self.raw_tick[0], '%Y-%m-%dT%H:%M:%S')
        self.bar_open   = float(self.raw_tick[2])
        self.bar_high   = float (self.raw_tick[3])
        self.bar_low    = float(self.raw_tick[4])
        self.bar_close  = float(self.raw_tick[5])
        self.bar_volume =  int(self.raw_tick[6])
        self.oi         = int(self.raw_tick[7])
        self.ttq        = int(self.raw_tick[8])
        
    def __repr__(self):
        bar_tuple = namedtuple('bar_feed', field_names= bar_fields)
        attributes = list(map(lambda x: getattr(self , x), bar_fields))
        return str(bar_tuple(*attributes))    
    

@dataclass
class greek_feed:
    raw_tick : List[any] = field(init = True , repr = False )
    symbol :str = field(init = True)
    
    def __post_init__(self):
        self.symbol_id  = int(self.raw_tick[0])
        self.timestamp  = datetime.strptime(self.raw_tick[1], '%Y-%m-%dT%H:%M:%S')
        self.iv     = float(self.raw_tick[2])
        self.delta  = float (self.raw_tick[3])
        self.theta  = float(self.raw_tick[4])
        self.gamma  = float(self.raw_tick[5])
        self.vega   =  float(self.raw_tick[6])
        self.rho    = float(self.raw_tick[7])

    def __repr__(self):
        greek_tuple = namedtuple('greek_feed', field_names= greek_fields)
        attributes = list(map(lambda x: getattr(self , x), greek_fields))
        return str(greek_tuple(*attributes))    
        
       
def get_symbol_id(segments , url , user_id , password , cache_dir):
    cache = {}
    params =  { 'user': user_id ,'password' : password,
                'allexpiry':'false', 'csv':'true','csvheader':'true'}
    for segment in segments:
        params.update({'segment': f'{segment}'.lower(),})
        response = requests.get(url, params= params).text
        df = pd.read_csv(StringIO(response) , dtype='unicode' )
        df = df[['symbolid', 'symbol']]
        df.set_index('symbolid' , inplace = True )
        seg = df.to_dict()['symbol']
        cache.update(seg)
    os.makedirs(cache_dir , exist_ok=True )
    with open( f"{cache_dir}/sym_cache_{dt.now().strftime('%d%m%y')}.pkl" , 'wb') as pkl:
        pickle.dump(cache , pkl)


def cache_symbol_id(username , password , td_obj ):
    if os.name == 'nt':
        sym_cache_dir = '/'.join(__file__.split('\\')[:-1]) + '/cache/sym_cache/'
    else: 
        sym_cache_dir = '/'.join(__file__.split('/')[:-1]) + '/cache/sym_cache/'
    url = 'https://api.truedata.in/getAllSymbols?'
    segments = ['EQ', 'FO', 'MCX', 'CDS', 'IN' , 'BSEEQ' , 'BSEFO']
    if not os.path.exists(sym_cache_dir) or not os.path.exists(f'{sym_cache_dir}/sym_cache_{dt.now().strftime("%d%m%y")}.pkl'):
        td_obj.logger.warning(f'{Style.NORMAL}{Fore.BLUE}please wait two minute to download master contracts for today{Style.RESET_ALL}')
        shutil.rmtree(sym_cache_dir, ignore_errors=True, onerror=None)
        get_symbol_id(segments , url , username , password , sym_cache_dir )
    with open( f"{sym_cache_dir}/sym_cache_{dt.now().strftime('%d%m%y')}.pkl" , 'rb') as pkl:
        symbols_map = pickle.load(pkl)
    return symbols_map

def remove_all_cache():
    if os.name == 'nt':
        cache_dir = '/'.join(__file__.split('\\')[:-1]) + '/cache'
    else: 
        cache_dir = '/'.join(__file__.split('/')[:-1]) + '/cache'
    shutil.rmtree(cache_dir, ignore_errors=True, onerror=None)
    

def get_atm(username , password , symbol , expiry):
    url = "https://api.truedata.in/getATMStrike?"
    expiry = expiry.strftime("%Y%m%d")
    params = {"user": username, "password": password, "symbol": symbol , "expiry": expiry}
    response = requests.get(url , params= params)
    return response.json()['Records'].values()


from .utils import historical_decorator, access_token_decorator, check_response
from io import StringIO
from datetime import datetime, timedelta
from datetime import datetime as dt
from logging import Logger
from threading import RLock
from colorama import Style, Fore
from typing import List, Dict
import os
import requests
import pandas as pd
import lz4.block 
import struct

class HistoricalREST:

    def __init__(self, login_id: str, password: str, url: str, logger: Logger):  # NO PORT, broker token needed from now on
        self.login_id = login_id
        self.password = password
        self.url = url
        self.logger = logger
        self.thread_lock = RLock()
        self.access_token = None
        self.bhavcopy_last_completed = None
        self.access_token_expiry_time = None
        try:
            self.hist_login()
        except Exception as e:
            self.logger.error(f"Failed to connect REST historical API -> {type(e)} = {e}")

    def get_new_token(self ):
        url_auth = "https://auth.truedata.in/token"
        payload = f"username={self.login_id}&password={self.password}&grant_type=password"
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        token_json = requests.request("POST", url_auth, headers=headers, data=payload).json()
        return token_json

    def hist_login(self):
        token_json = self.get_new_token()
        try:
            if token_json['access_token']:
                self.access_token = token_json['access_token']
                self.access_token_expiry_time = datetime.now() + timedelta(seconds=token_json['expires_in'] - 15)  # 15 seconds is random buffer time
                self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Connected successfully to TrueData Historical Data Service... {Style.RESET_ALL}")
        except Exception as e:
            self.logger.error(f"Failed to connect -> {token_json['error_description']}{type(e)} = {e}")
            self.access_token = None

    # noinspection PyUnusedLocal
    @access_token_decorator
    @historical_decorator
    def get_n_historic_bars(self, contract, end_time, no_of_bars, bar_size, bidask=False):
        end_point = 'getlastnbars'
        try:
            headers = {'Authorization': f'Bearer {self.access_token}'}
            encoded_payload = { 'symbol': contract, 'interval': bar_size, 'response': 'csv', 'bidask': 0 ,'comp': 'true',}
            if bar_size == 'tick':
                encoded_payload['nticks'] = no_of_bars
                end_point = 'getlastnticks'
                if bidask:
                    encoded_payload['bidask'] = 1
            else:  # Not ticks
                encoded_payload['nbars'] = no_of_bars
            with self.thread_lock:
                url = f"{self.url}/{end_point}"
                response = requests.get(url, headers=headers, params=encoded_payload)
                check_response(response)
                hist_data = HistoricalREST.decompress_data(response.content)
                hist_data = self.parse_data(hist_data)
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None
        return hist_data 
    
    @access_token_decorator
    @historical_decorator
    def get_historic_data(self, contract, end_time, start_time, bar_size, delivery = False , bidask=False ):
        end_point = 'getbars'
        try:
            encoded_payload = {'symbol': contract,'interval': bar_size,'response': 'csv','comp': 'true',}
            encoded_payload.update({'delivery' : 'true'}) if delivery and bar_size =="eod" else 0
            unencoded_payload = [f"from={start_time}", f"to={end_time}"]
            unencoded_payload = "&".join(unencoded_payload)
            if bar_size == 'tick':
                encoded_payload['bidask'] = 0
                end_point = 'getticks'
                if bidask:
                    encoded_payload['bidask'] = 1
            headers = { 'Authorization': f'Bearer {self.access_token}'}
            with self.thread_lock:
                url = f"{self.url}/{end_point}?{unencoded_payload}"
                response = requests.get(url, headers=headers, params=encoded_payload)
                check_response(response)
                hist_data = HistoricalREST.decompress_data(response.content)
                hist_data = self.parse_data(hist_data)
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None
        return hist_data

    def get_gainers_losers(self , segment, topn , gainers  ):
        source_string = "gettopngainers" if gainers else "gettopnlosers"
        headers = {'Authorization': f'Bearer {self.access_token}'}
        payload = {'segment': segment , 'topn': topn , 'response':'csv' }
        url = f"{self.url}/{source_string}?"
        try:
            with self.thread_lock:
                response = requests.get(url, headers=headers , params=payload )
                check_response(response)
                data = response.text
            data = self.parse_data(data)
        except Exception as e:
            self.logger.error(f"{type(e)} -> No match found for this segment {segment}")
            return None
        return data

    def parse_data(self, data):
        df = pd.read_csv(StringIO(data) , index_col = None)
        df.timestamp = pd.to_datetime(df.timestamp)
        return df

    def bhavcopy_status(self, segment: str):
        url = f'{self.url}/getbhavcopystatus?segment={segment}&response=csv'
        headers = {'Authorization': f'Bearer {self.access_token}'}
        response = requests.request("GET", url, headers=headers)
        status = response.text.strip().split('\r\n')[-1].split(',')
        assert segment == status[0]
        self.bhavcopy_last_completed = datetime.strptime(status[1], "%Y-%m-%dT%H:%M:%S")

    @access_token_decorator
    def bhavcopy(self, segment: str, date: datetime) -> List[Dict]:
        try:
            self.bhavcopy_status(segment)
            if date > self.bhavcopy_last_completed:
                self.logger.error(f"{Style.BRIGHT}{Fore.RED}No complete bhavcopy found for requested date."
                                    f" Last available for {self.bhavcopy_last_completed.strftime('%Y-%m-%d %H:%M:%S')}.{Style.RESET_ALL}")
                return []
            url_bhavcopy = f"{self.url}/getbhavcopy?segment={segment}&date={date.strftime('%Y-%m-%d')}&response=csv"
            payload = { 'comp' : 'true' }
            headers = {'Authorization': f'Bearer {self.access_token}'}
            response = requests.request( "GET", url_bhavcopy, headers=headers, params=payload )
            check_response(response)
            response = HistoricalREST.decompress_data( response.content )
            data = self.parse_data(response)
            return data 
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")

    @staticmethod
    def decompress_data(data):
        uncom_length = struct.unpack('<I', data[:4])[0]
        com_length = struct.unpack('<I', data[4:8])[0]
        dc = lz4.block.decompress( data[8:], uncom_length ) if com_length != uncom_length else data[8:]
        return dc.decode()



from . import websocket
from .tests.defaults import MIN_PYTHON_MINOR, MIN_PYTHON_MAJOR
from .history import TD_hist
from .analytics import TD_analytics
from .websocket import TD_live

__version__ = '6.0.0-beta.3'



from colorama import Style, Fore    

class TDLiveCalcError(Exception):
    def __str__(self):
        return f"{Style.BRIGHT}{Fore.RED}Something's wrong with the live calculations- {self.args[0]}{Style.RESET_ALL}"

class TDLiveDataError(Exception):
    def __str__(self):
        return f"{Style.BRIGHT}{Fore.RED}Something's wrong with the live data- {self.args[0]}{Style.RESET_ALL}"

class TDInvalidRequestError(Exception):
    def __str__(self):
        return f"{Style.BRIGHT}{Fore.RED}Invalid request ({self.args[0]}){Style.RESET_ALL}"



from io import StringIO
from copy import deepcopy
from datetime import datetime as dt
from datetime import time as t
from .constants import chain_live_update_column , chain_min_update_column , chain_columns , chain_fields_from_live_dict  , chain_fields_from_min_dict , chain_greek_fields 
import pandas as pd
import requests
import re
import numpy as np
import time



class OptionChain:

    chain_url ='https://api.truedata.in/getOptionChain?'

    def __init__(self, TD_OBJ , symbol ,expiry , chain_length ,atm , strike_step , bid_ask , subs , greek_status):
        self.TD_OBJ = TD_OBJ
        self.symbol = symbol
        self.expiry = expiry
        self.chain_length = chain_length
        self.atm = atm
        self.strike_step = strike_step
        self.subscription = subs
        self.bid_ask = bid_ask
        self.greek_status = greek_status
        self.option_symbols = self.get_option_symbols( atm = self.atm)
        self.chain_dataframe =  self.init_dataframe()
        self.chain_status = False

    def get_option_symbols(self , atm ):
        expiry = self.expiry.strftime('%y%m%d')
        start_strike = atm - self.strike_step * int(self.chain_length / 2 )
        end_strike = atm + self.strike_step * int( self.chain_length / 2)
        req_strikes = np.arange( start_strike , end_strike  , self.strike_step )
        if isinstance( self.strike_step , float ) :
            req_strikes = list(map(lambda x: int(x) if x.is_integer() else x , req_strikes))
        symbols = list(map(lambda x: f'{self.symbol}{expiry}{x}CE', req_strikes))
        symbols.extend(list(map(lambda x: f'{self.symbol}{expiry}{x}PE', req_strikes )))
        return symbols

    def init_dataframe(self):
        my_chain_columns = deepcopy(chain_columns)
        my_chain_columns.extend(chain_greek_fields) if self.greek_status else chain_columns
        df = pd.DataFrame(columns= my_chain_columns) if self.subscription == 'tick' else pd.DataFrame(columns= my_chain_columns[:-4])
        df.symbols = self.option_symbols
        df.strike = df.symbols.apply(lambda x: str(re.search(r"\d+(\.\d+)?", x ).group(0) )[6:])
        df.type = df.symbols.apply(lambda x: re.findall('\D+' , x )[-1] )
        df.set_index('symbols' , inplace=True)
        df.sort_values(by=['strike' , 'type'] , inplace=True)
        return df

    def initial_update_from_touchline(self , data_dict , equivalent_fields , column):
        for symbol in self.option_symbols:
            try:
                tick_data = data_dict[symbol]
                live_data_attributes = list(map(lambda x: getattr(tick_data , x), equivalent_fields))
                self.chain_dataframe.loc[symbol , column] = live_data_attributes
            except TypeError:
                self.chain_dataframe.loc[symbol , column] = [ np.NaN for i in range(14)]
            except KeyError:
                pass
       
    def update_chain(self ):
        self.chain_status = True
        data_dict = self.TD_OBJ.live_data if self.subscription == 'tick' else self.TD_OBJ.one_min_live_data 
        chain_equivalent_columns = chain_live_update_column if self.subscription == 'tick' else chain_min_update_column
        equivalent_fields = chain_fields_from_live_dict if self.subscription == 'tick' else chain_fields_from_min_dict
        self.initial_update_from_touchline( data_dict = data_dict , equivalent_fields = equivalent_fields , column = chain_equivalent_columns )
        while self.chain_status:
            for symbol in self.option_symbols:
                try:
                    tick_data = data_dict[symbol]
                    live_data_attributes = list(map(lambda x: getattr(tick_data , x), equivalent_fields))
                    self.chain_dataframe.loc[symbol , chain_equivalent_columns] = live_data_attributes
                    if self.greek_status and symbol in self.TD_OBJ.greek_data.keys():
                        greek_data = self.TD_OBJ.greek_data[symbol]
                        greek_data_attributes = list(map(lambda x: getattr(greek_data , x), chain_greek_fields))
                        self.chain_dataframe.loc[symbol , chain_greek_fields] = greek_data_attributes
                except KeyError:
                    pass
            time.sleep(1)

    def get_option_chain(self):
        df = self.chain_dataframe.copy()
        # print(df)''
        df.dropna(subset= ['ltt'] , inplace=True)
        df = df.astype({'ltq' : int , 'volume': int })
        if not self.bid_ask and 'bid' in df.columns:
            df = df.drop(['bid' , 'bid_qty' , 'ask' , 'ask_qty'] , axis= 1  )
        return df

    def stop_option_chain(self):
        self.chain_status = False
        self.TD_OBJ.stop_live_data(self.option_symbols)




from .utils import TDHistoricDataError
from .Historical_REST import HistoricalREST 
from datetime import datetime
from dateutil.relativedelta import relativedelta                                # type: ignore
from typing import List, Dict
import logging


class TD_hist:

    def __init__(   self, login_id, password, log_level=logging.WARNING,
                    log_handler=None, log_format=None, hist_url='https://history.truedata.in' ):
        self.historical_datasource = None
        self.login_id = login_id
        self.password = password
        self.hist_url = hist_url
        self.set_custom_log(log_level , log_handler , log_format)
        self.connect()

    def connect(self):
        self.historical_datasource = HistoricalREST(self.login_id, self.password, self.hist_url, self.logger)
        
    def set_custom_log(self , log_level , log_handler , log_format ):
        if log_format is None:
            log_format = "(%(asctime)s) %(levelname)s :: %(message)s (PID:%(process)d Thread:%(thread)d)"
        if log_handler is None:
            log_formatter = logging.Formatter(log_format)
            self.log_handler = logging.StreamHandler()
            self.log_handler.setLevel(log_level)
            self.log_handler.setFormatter(log_formatter)
        else:
            self.log_handler = log_handler
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(self.log_handler)
        self.logger.setLevel(self.log_handler.level)
        self.logger.debug("Logger ready...")

    @staticmethod
    def truedata_duration_map(regular_format, end_date):
        duration_units = regular_format.split()[1].upper()
        if len(duration_units) > 1:
            raise TDHistoricDataError("Misconfigured duration argument")
        duration_size = int(regular_format.split()[0])
        if duration_units == 'D':
            return (end_date - relativedelta(days=duration_size - 1)).date()
        elif duration_units == 'W':
            return (end_date - relativedelta(weeks=duration_size)).date()
        elif duration_units == 'M':
            return (end_date - relativedelta(months=duration_size)).date()
        elif duration_units == 'Y':
            return (end_date - relativedelta(years=duration_size)).date()

    def get_historic_data(  self, contract, end_time=None, duration=None,
                            start_time=None, bar_size="1 min", bidask=False, delivery=False):
        if delivery and not bar_size.lower() == "eod":
            delivery = False
        if start_time is not None and duration is None:
            return self.get_historical_data_from_start_time(contract=contract, end_time=end_time,
                                                            start_time=start_time, bar_size=bar_size,
                                                            bidask=bidask, delivery = delivery)
        else:
            return self.get_historical_data_from_duration(contract=contract, end_time=end_time,
                                                          duration=duration, bar_size=bar_size,
                                                        bidask=bidask, delivery = delivery)

    def get_n_historical_bars(  self, contract, end_time: datetime = None, 
                                no_of_bars: int = 1, bar_size="1 min", bidask=False):
        if end_time is None:
            end_time = datetime.today()
        end_time = end_time.strftime('%y%m%dT%H:%M:%S')    # This is the request format
        hist_data = self.historical_datasource.get_n_historic_bars(contract, end_time, no_of_bars,
                                                                   bar_size, bidask=bidask)
        return hist_data

    def get_historical_data_from_duration(  self, contract, delivery, end_time: datetime = None, duration=None,
                                            bar_size="1 min", bidask=False ):
        if duration is None:
            duration = "1 D"
        if end_time is None:
            end_time = datetime.today()
        start_time = self.truedata_duration_map(duration, end_time)
        end_time = end_time.strftime('%y%m%d') + 'T23:59:59'    # This is the request format
        start_time = start_time.strftime('%y%m%d') + 'T00:00:00'    # This is the request format
        hist_data = self.historical_datasource.get_historic_data(   contract, end_time, start_time, bar_size, 
                                                                    bidask=bidask ,delivery = delivery )
        return hist_data

    def get_historical_data_from_start_time(self, contract,delivery, end_time: datetime = None,
                                            start_time: datetime = None, bar_size="1 min", bidask=False):
        if end_time is None:
            end_time = datetime.today().replace(hour=23, minute=59, second=59)
        if start_time is None:
            start_time = datetime.today().replace(hour=0, minute=0, second=0)
        end_time = end_time.strftime('%y%m%dT%H:%M:%S')    # This is the request format
        start_time = start_time.strftime('%y%m%dT%H:%M:%S')    # This is the request format
        hist_data = self.historical_datasource.get_historic_data(   contract, end_time, start_time, bar_size, 
                                                                    bidask=bidask , delivery = delivery )
        return hist_data

    def get_bhavcopy(self, segment: str, date: datetime = None) -> List[Dict]:
        if date is None:
            date = datetime.now().replace(hour=0, minute=0, second=0)
        return self.historical_datasource.bhavcopy(segment, date)

    def get_gainers(self , segment , topn , df_style = True ):
        topn = 10 if topn is None else topn
        return self.historical_datasource.get_gainers_losers(segment, topn , gainers = True  )

    def get_losers(self , segment , topn , df_style = True ):
        topn = 10 if topn is None else topn
        return self.historical_datasource.get_gainers_losers(segment , topn , gainers = False  )


from datetime import datetime as dt
from functools import wraps


def access_token_decorator(func):
    @wraps(func)
    def dec_helper(obj, *args, **kwargs):
        if obj.access_token_expiry_time < dt.now():
            obj.hist_login()
        return func(obj, *args, **kwargs)
    return dec_helper


from .TD_ws import LiveClient
from .TD_chain import OptionChain
from .utils import remove_all_cache , cache_symbol_id , get_atm
from threading import Thread
from typing import Callable
from datetime import datetime
from colorama import Style, Fore                                                # type: ignore
import time
import json
import logging
import traceback

class TD_live:
    def __init__(self, login_id, password, url='push.truedata.in', live_port=8082, log_level=logging.WARNING,
                 log_handler=None, log_format=None, full_feed = False, dry_run = False):
        self.login_id = login_id
        self.password = password
        self.live_url = url
        self.live_port = live_port
        self.full_feed = full_feed
        self.dry_run = dry_run
        self.live_websocket = None
        self.set_custom_log(log_level , log_handler , log_format)
        if live_port is None:
            self.connect_live = False
        else:
            self.connect_live = True
        self.live_data = {}
        self.one_min_live_data = {}
        self.five_min_live_data = {}
        self.touchline_data = {}
        self.greek_data = {}
        self.connect()

    def set_custom_log(self , log_level , log_handler , log_format ):
        if log_format is None:
            log_format = "(%(asctime)s) %(levelname)s :: %(message)s (PID:%(process)d Thread:%(thread)d)"
        if log_handler is None:
            log_formatter = logging.Formatter(log_format)
            self.log_handler = logging.StreamHandler()
            self.log_handler.setLevel(log_level)
            self.log_handler.setFormatter(log_formatter)
        else:
            self.log_handler = log_handler
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(self.log_handler)
        self.logger.setLevel(self.log_handler.level)
        self.logger.debug("Logger ready...")
  
    def connect(self):
        broker_append = ''
        if self.dry_run:
            remove_all_cache()
        self.symbol_id_map_dict = cache_symbol_id(self.login_id , self.password  , self ) if self.full_feed else 0
        if self.connect_live:
            self.live_websocket = LiveClient(self, f"wss://{self.live_url}:{self.live_port}?user={self.login_id}&password={self.password}{broker_append}" )
            t = Thread(target=self.connect_thread, args=(), daemon=True)
            t.start()
        while self.connect_live and self.live_websocket.subscription_type == '':
            time.sleep(1)

    def connect_thread(self):
        self.live_websocket.run_forever(ping_interval=10, ping_timeout=5)
        while not self.live_websocket.disconnect_flag:
            self.logger.info(f"{Style.BRIGHT}{Fore.RED}Connection dropped due to no network , Attempting reconnect @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
            self.live_websocket.reconnect()
            time.sleep(10)
        self.logger.debug('Goodbye (properly) !!')

    def disconnect(self):
        if self.connect_live:
            self.live_websocket.disconnect_flag = True
            self.live_websocket.close()
            self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Disconnected from Real Time Data WebSocket Connection !{Style.RESET_ALL}")

    def start_live_data(self, symbols , restart_flag = False):  # TODO: Prevent reuse of req_ids
        if restart_flag:
            self.live_websocket.send(f'{{"method": "addsymbol", "symbols": {json.dumps(symbols)}}}')
            return True
        symbols_to_call = []
        symbols = list(set(symbols))
        for symbol in symbols:
            symbol = symbol.upper()
            symbols_to_call.append(symbol)
        if len(symbols_to_call) > 0:
            self.live_websocket.send(f'{{"method": "addsymbol", "symbols": {json.dumps(symbols_to_call)}}}')
        return True
        
    def stop_live_data(self, symbols):  
        self.live_websocket.send(f'{{"method": "removesymbol", "symbols": {json.dumps(symbols)}}}')

    def trade_callback(self, func: Callable):
        self.logger.info(f"Defining {func} as trade_callback...")
        self.live_websocket.trade_callback = func

    def clear_trade_callback(self):
        self.logger.info(f"Clearing trade_callback...")
        self.live_websocket.trade_callback = None

    def bidask_callback(self, func: Callable):
        self.logger.info(f"Defining bidask_callback...")
        self.live_websocket.bidask_callback = func

    def clear_bidask_callback(self):
        self.logger.info(f"Clearing bidask_callback...")
        self.live_websocket.bidask_callback = None

    def one_min_bar_callback(self, func: Callable):
        self.logger.info(f"Defining one min bar_callback...")
        self.live_websocket.one_min_bar_callback = func

    def clear_one_min_bar_callback(self):
        self.logger.info(f"Clearing one min bar_callback...")
        self.live_websocket.one_min_bar_callback = None

    def five_min_bar_callback(self, func: Callable) :
        self.logger.info(f"Defining five min bar_callback...")
        self.live_websocket.five_min_bar_callback = func

    def clear_five_min_bar_callback(self):
        self.logger.info(f"Clearing five min bar_callback...")
        self.live_websocket.five_min_bar_callback = None

    def full_feed_trade_callback(self , func: Callable ):
        self.logger.info(f"Defining full feed tick_callback...")
        self.live_websocket.full_feed_trade_callback = func

    def clear_full_feed_trade_callback(self):
        self.logger.info(f"Clearing full feed trade_callback...")
        self.live_websocket.full_feed_trade_callback = None
    
    def full_feed_bar_callback(self , func: Callable ):
        self.logger.info(f"Defining full feed bar_callback...")
        self.live_websocket.full_feed_bar_callback = func

    def clear_full_feed_bar_callback(self):
        self.logger.info(f"Clearing full feed bar_callback...")
        self.live_websocket.full_feed_bar_callback = None

    def greek_callback(self , func: Callable ):
        self.logger.info(f"Defining greek feed callback...")
        self.live_websocket.greek_callback = func

    def clear_greek_callback(self):
        self.logger.info(f"Clearing greek feed callback...")
        self.live_websocket.greek_callback = None

    # def get_touchline(self):
    #     self.live_websocket.send(f'{{"method": "touchline"}}')

    def start_option_chain( self , symbol , expiry , chain_length = None , bid_ask = False , greek = False ):
        chain_length = 10 if chain_length is None else chain_length  # setting default value for chain length
        try:
            atm , strike_step  = get_atm(self.login_id , self.password , symbol , expiry  )
            chain = OptionChain(self , symbol, expiry , chain_length , atm , strike_step , bid_ask , self.live_websocket.subscription_type , greek )
        except Exception as e:
            traceback.print_exc()
            print(f'Please check symbol: {symbol} and its expiry: {expiry.date()}')
            exit()
        self.start_live_data(chain.option_symbols)
        time.sleep(2)
        option_thread = Thread(target=chain.update_chain, args=( ), daemon=True)
        option_thread.start()
        return chain



from colorama import Style, Fore         # type: ignore
from functools import wraps
from datetime import datetime
import base64

class TDHistoricDataError(Exception):
    def __str__(self):
        return f"{Style.BRIGHT}{Fore.RED}Something's wrong with the historical data- {self.args[0]}{Style.RESET_ALL}"

class TooManyRequestsError(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self ):
        return self.msg

class NotFoundError(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self ):
        return self.msg

def historical_decorator(func):
    @wraps(func)
    def dec_helper(obj, contract, end_time, start_time, bar_size, delivery = False, bidask=False):
        if bar_size.lower() == 'tick' or bar_size.lower() == 'ticks':
            bar_size = 'tick'            
        elif bar_size.lower() == 'eod' or bar_size.lower() == 'week' or bar_size.lower() == 'month':
            bar_size = bar_size.lower()
        else:
            bar_size = bar_size.replace(' ', '')
            if bar_size[-1] == 's':
                bar_size = bar_size[:-1]
        if delivery:
            return func(obj, contract, end_time, start_time, bar_size, delivery, bidask)
        return func(obj, contract, end_time, start_time, bar_size, bidask = bidask)
    return dec_helper


def access_token_decorator(func):
    @wraps(func)
    def dec_helper(obj, *args, **kwargs):
        if obj.access_token_expiry_time < datetime.now():
            obj.hist_login()
        return func(obj, *args, **kwargs)
    return dec_helper


def check_response(response):
    if response.status_code == 429:
        error = base64.b64decode(response.content).decode('utf-8')
        raise TooManyRequestsError(error)
    elif response.status_code == 404:
        error = response.text
        raise NotFoundError(error)
    elif response.status_code == 200:
        return None
    else:
        error = response.text
        raise Exception(error)


from websocket import WebSocketApp
from threading import Thread
from datetime import datetime
from dateutil.relativedelta import relativedelta            # type: ignore
from copy import deepcopy
from .utils import full_feed , bidask_feed , bar_feed , greek_feed , tick_feed , min_feed, TouchlineData
from colorama import Style, Fore                            # type: ignore
import time
import json
import traceback

class LiveClient(WebSocketApp):

    def __init__(self, parent_app, url, *args):
        WebSocketApp.__init__(self, url, on_open=self.ws_open, on_error=self.ws_error, on_message=self.on_msg_func, on_close=self.ws_close, *args)
        self.segments = []
        self.max_symbols = 0
        self.remaining_symbols = 0
        self.valid_until = ''
        self.subscription_type = ''
        self.confirm_heartbeats = 1
        self.store_last_n_heartbeats = self.confirm_heartbeats + 7
        self.heartbeat_interval = 5
        self.heartbeat_buffer = 0.5
        self.parent_app = parent_app
        self.symbol_id_map = {}
        time_of_creation = datetime.now()
        self.last_n_heartbeat = [time_of_creation - relativedelta(seconds=i * self.heartbeat_interval) for i in range(-self.store_last_n_heartbeats, 0)]
        self.logger = self.parent_app.logger
        self.disconnect_flag = False
        self.heartbeat_check_thread = Thread(target=self.check_heartbeat, daemon=True)
        self.trade_callback = None
        self.full_feed_trade_callback = None
        self.full_feed_bar_callback = None
        self.bidask_callback = None
        self.one_min_bar_callback = None
        self.five_min_bar_callback = None
        self.greek_callback = None
        self.subs = ''


    def check_connection(self):
        base_heartbeat = self.last_n_heartbeat[-self.confirm_heartbeats]
        check_time = datetime.now()
        time_diff = check_time - base_heartbeat
        is_connected = time_diff.total_seconds() > ((self.heartbeat_interval + self.heartbeat_buffer) * self.confirm_heartbeats)  # 3 > 5 + 0.5
        return is_connected

    def check_heartbeat(self):
        while True:
            time.sleep(self.heartbeat_interval)
            if self.disconnect_flag:
                self.logger.info(f"{Fore.WHITE}Removing hand from the pulse...{Style.RESET_ALL}")
                break
            if self.check_connection():
                self.logger.debug(f"{Style.BRIGHT}{Fore.RED}Failed Heartbeat @ {datetime.now()} because of last at {self.last_n_heartbeat[-self.confirm_heartbeats]}{Style.RESET_ALL}")
                self.logger.info(f"{Style.BRIGHT}{Fore.RED}Attempting reconnect @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
                restart_successful = self.reconnect()
                if restart_successful:
                    self.logger.info(f"{Style.BRIGHT}{Fore.GREEN}Successful restart @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
                    time.sleep((self.heartbeat_interval + self.heartbeat_buffer))
                    recover_start, recover_end = self.get_largest_diff(self.last_n_heartbeat)
                    self.recover_from_time_missed(recover_start, recover_end)
                else:
                    self.logger.info(f"{Style.BRIGHT}{Fore.RED}Failed restart @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")

    @staticmethod
    def get_largest_diff(time_series):
        big_li = deepcopy(time_series.pop(0))
        small_li = deepcopy(time_series.pop())
        diffs = [i[0]-i[1] for i in zip(big_li, small_li)]
        start_gap_index = max(range(len(diffs)), key=lambda i: diffs[i])
        return time_series[start_gap_index], time_series[start_gap_index+1]

    def recover_from_time_missed(self, start_time, end_time):
        self.logger.info(f"{Style.BRIGHT}{Fore.YELLOW}Initiating recovery from {Fore.GREEN}{start_time}{Fore.YELLOW} till {Fore.GREEN}{end_time}{Fore.YELLOW} "
                         f"which are last green heartbeats from server...{Style.RESET_ALL}")

    def reconnect(self):
        self.close()
        time.sleep(2)
        t = Thread(target=self.run_forever, daemon=True)
        t.start()
        is_td_connected = False
        while not is_td_connected:
            time.sleep(self.heartbeat_interval + self.heartbeat_buffer)
            is_td_connected = self.check_connection()
        return is_td_connected

    def on_msg_func(self, *args):
        try:
            message = args[-1]
            msg = json.loads(message)
            msg_keys = msg.keys()
            # print(msg) 
            if 'message' in msg_keys:
                self.handle_message_data(msg)
            if self.parent_app.full_feed:
                if 'trade' in msg_keys :
                    feed_tick = full_feed( raw_tick=msg['trade'] ,symbol=self.parent_app.symbol_id_map_dict[msg['trade'][0]] )
                    self.full_feed_trade_callback( feed_tick ) if self.full_feed_trade_callback else None   
                elif ('bidask' in msg_keys or 'bidaskL2' in msg_keys )  :
                    if 'bidask' in msg_keys:
                        bid_ask = bidask_feed( raw_tick=msg['bidask'] , symbol=self.parent_app.symbol_id_map_dict[msg['bidask'][0]] , level= "L1" )
                    else:
                        bid_ask = bidask_feed( raw_tick=msg['bidaskL2'] , symbol=self.parent_app.symbol_id_map_dict[msg['bidaskL2'][0]] , level= "L2" )
                    self.bidask_callback( bid_ask ) if self.bidask_callback else None  
                elif 'interval' in msg_keys  :
                    bar_data = map(lambda x : bar_feed(x , self.parent_app.symbol_id_map_dict[ str( x[1]) ]), msg["data"]) 
                    for bar in list(bar_data):
                        self.full_feed_bar_callback ( bar ) if self.full_feed_bar_callback else None
                elif 'greeks' in msg_keys:
                    greeks = greek_feed( raw_tick=msg['greeks'] , symbol= self.parent_app.symbol_id_map_dict[msg['greeks'][0]] )
                    self.greek_callback(greeks) if self.greek_callback else None
            else:
                if 'trade' in msg_keys:
                    trade_tick = self.update_trade_data( msg['trade'] )
                    self.trade_callback(trade_tick) if self.trade_callback else None
                elif ('bidask' in msg_keys or 'bidaskL2' in msg_keys ) :
                    if 'bidask' in msg_keys:
                        bid_ask = bidask_feed( raw_tick=msg['bidask'] , symbol=self.symbol_id_map[msg['bidask'][0]] , level = "L1" )
                    else:
                        bid_ask = bidask_feed( raw_tick=msg['bidaskL2'] , symbol=self.symbol_id_map[msg['bidask'][0]] , level = "L2" )
                    self.bidask_callback( bid_ask ) if self.bidask_callback else None
                elif 'bar1min' in msg_keys:
                    self.update_bar_data( msg['bar1min'] , bar_type= '1min' )
                elif 'bar5min' in msg_keys:
                    self.update_bar_data( msg['bar5min'] , bar_type= '5min')
                elif 'greeks' in msg_keys:
                    greeks = greek_feed( raw_tick=msg['greeks'] , symbol=self.symbol_id_map[msg['greeks'][0]])
                    self.parent_app.greek_data[greeks.symbol] = greeks
                    self.greek_callback(greeks) if self.greek_callback else None
        except KeyError as e:
            self.logger.warning(f"{Style.BRIGHT}{Fore.RED}Symbol id - {e.args[0]} is not availavble in the masters and cannot be mapped.{Style.RESET_ALL}")
        except Exception :
            traceback.print_exc()
            exit()

    def handle_message_data(self, msg):
        if msg['success']:
            if msg['message'] == 'HeartBeat':
                self.handle_heartbeat(msg['timestamp'])
            elif msg['message'] == 'TrueData Real Time Data Service':  # Connection success message
                self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Connected successfully to {msg['message']}... {Style.RESET_ALL}")
                self.subscription_type = msg['subscription']
                self.subs = self.subscription_type.split('+')
                # print(self.subscription_type)
            elif msg['message'] in ['symbols added', 'touchline']:
                self.update_symbols_data(msg['symbollist'])
            elif msg['message'] == 'symbols removed':
                self.remove_symbols(msg['symbollist'])
            elif msg['message'] == 'marketstatus':
                self.logger.debug(f"Market status message -> {msg['data']}") 
        else:
            self.logger.error(f"{Style.BRIGHT}{Fore.RED}The request encountered an error - {msg['message']}{Style.RESET_ALL}")

    def handle_heartbeat(self, server_timestamp):
        self.logger.debug(f'Server heartbeat received at {server_timestamp}')
        timestamp = datetime.strptime(server_timestamp, "%Y-%m-%dT%H:%M:%S.%f")
        self.last_n_heartbeat = self.last_n_heartbeat[1:]
        self.last_n_heartbeat.append(timestamp)

    def remove_symbols(self, symbols):
        for symbol in symbols:
            symb = symbol.split(':')[0]
            self.parent_app.touchline_data.pop(symb, None) 
            self.parent_app.live_data.pop(symb, None) 
            self.parent_app.one_min_live_data.pop(symb, None) 
            self.parent_app.five_min_live_data.pop(symb, None) 

    def update_symbols_data(self, symbols_data):
        for symbol_data in symbols_data:
            symbol = symbol_data[0]
            self.symbol_id_map[symbol_data[1]] = symbol
            self.parent_app.touchline_data[symbol] = TouchlineData(raw_data= symbol_data)
            if 'tick' in self.subs:
                self.parent_app.live_data[symbol] = tick_feed(touchline = self.parent_app.touchline_data[symbol])
            if '1min' in self.subs:
                self.parent_app.one_min_live_data[symbol] = min_feed(touchline = self.parent_app.touchline_data[symbol])
            if '5min' in self.subs:
                self.parent_app.five_min_live_data[symbol] = min_feed(touchline = self.parent_app.touchline_data[symbol])
            self.logger.info(f"Updated touchline data for {symbol} and related live data objects...")

    def update_trade_data(self, trade_tick):
        try:
            symbol = self.symbol_id_map[trade_tick[0]]
            self.parent_app.live_data[symbol].timestamp = datetime.strptime(trade_tick[1], '%Y-%m-%dT%H:%M:%S')
            self.parent_app.live_data[symbol].ltp =  float(trade_tick[2])
            self.parent_app.live_data[symbol].ltq = float(trade_tick[3])
            self.parent_app.live_data[symbol].atp = float(trade_tick[4])
            self.parent_app.live_data[symbol].ttq = float(trade_tick[5])
            self.parent_app.live_data[symbol].day_open = float(trade_tick[6])
            self.parent_app.live_data[symbol].day_high = float(trade_tick[7])
            self.parent_app.live_data[symbol].day_low = float(trade_tick[8])
            self.parent_app.live_data[symbol].prev_day_close = float(trade_tick[9])
            self.parent_app.live_data[symbol].oi = int(trade_tick[10])
            self.parent_app.live_data[symbol].prev_day_oi = int(trade_tick[11])
            self.parent_app.live_data[symbol].turnover = float(trade_tick[12])
            self.parent_app.live_data[symbol].special_tag = str(trade_tick[13])
            self.parent_app.live_data[symbol].tick_seq = int(trade_tick[14])
            if len(trade_tick) > 15:
                self.parent_app.live_data[symbol].best_bid_price = float(trade_tick[15])
                self.parent_app.live_data[symbol].best_bid_qty = int(trade_tick[16])
                self.parent_app.live_data[symbol].best_ask_price = float(trade_tick[17])
                self.parent_app.live_data[symbol].best_ask_qty = int(trade_tick[18])
            return self.parent_app.live_data[symbol]
        except Exception as e:
            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with tick feed - {type(e)}{Style.RESET_ALL}')

    def update_bar_data(self, bar_data , bar_type ):
        try:
            symbol = self.symbol_id_map[bar_data[0]]
            obj_to_edit = self.parent_app.one_min_live_data[symbol] if bar_type == '1min' else self.parent_app.five_min_live_data[symbol] 
            obj_to_edit.timestamp = datetime.strptime(bar_data[1], '%Y-%m-%dT%H:%M:%S')
            obj_to_edit.open = float(bar_data[2])
            obj_to_edit.high = bar_high = float(bar_data[3])
            if bar_high > obj_to_edit.day_high:
                obj_to_edit.day_high = bar_high
            obj_to_edit.low = bar_low = float(bar_data[4])
            if bar_low < obj_to_edit.day_low:
                obj_to_edit.day_low = bar_low
            obj_to_edit.close = float(bar_data[5])
            obj_to_edit.volume = float(bar_data[6])
            obj_to_edit.oi = float(bar_data[7])
            if self.one_min_bar_callback and bar_type == '1min' :
                self.execute_callbacks(  obj_to_edit , '1min')  
            elif self.five_min_bar_callback and bar_type == '5min' :
                self.execute_callbacks(  obj_to_edit , '5min' ) 
        except Exception as e:
            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Bar feed encountered - {e}{Style.RESET_ALL}')

    def execute_callbacks(self , obj_to_edit , bar_type ):
        try:
            self.one_min_bar_callback( obj_to_edit) if bar_type == '1min' else self.five_min_bar_callback( obj_to_edit)
        except Exception as e:
            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with bar_callback - {type(e)} - {e}{Style.RESET_ALL}')

    def ws_error(self, *args):
        error = args[-1]
        if any(isinstance(error, conn_error) for conn_error in [ConnectionResetError, TimeoutError]):
            self.logger.error(f"Raising WS error = {error}")
            self.last_ping_tm = self.last_pong_tm = 0

    def ws_open(self, *args):
        self.last_ping_tm = time.time()
        self.sock.ping()
        self.sock.settimeout(15)
        if self.symbol_id_map:
            self.logger.info("connection dropped due to timeout. trying reconnecting.....")
            restart_symbols = list(self.symbol_id_map.values())
            self.logger.debug(f'{restart_symbols} needs resuming here.')
            self.sock.ping()
            time.sleep(3)
            self.parent_app.start_live_data(restart_symbols, restart_flag = True )
            self.logger.info(f'All streaming symbols have been resumed.')

    def ws_close(self, *args):
        self.sock.close() if self.sock is not None else 0
        self.sock = None
        if (self.last_ping_tm == 0 and self.last_pong_tm == 0) or (self.last_ping_tm > self.last_pong_tm) :
            self.logger.debug('DISCONNECTED FROM SERVER. RETRYING IN 5 SEC......')
            try:
                time.sleep(5)
                self.run_forever(ping_interval=10, ping_timeout=5)
            except Exception as e:
                self.logger.error(f'{type(e)} in reconnection => {e}')



from .decorators import access_token_decorator
from io import StringIO
from datetime import datetime as dt , timedelta , date
from threading import RLock
from colorama import Style, Fore
import logging
import requests
import pandas as pd
import lz4.block
import struct
import base64

class TD_analytics:

    def __init__(self, login_id, password, log_level = logging.WARNING , log_handler = None , log_format = None):  
        
        self.login_id = login_id
        self.password = password
        self.analytics_url = "https://analytics.truedata.in/api/"
        self.greeks_url = "https://greeks.truedata.in/api/"
        if log_format is None:
            log_format = "(%(asctime)s) %(levelname)s :: %(message)s (PID:%(process)d Thread:%(thread)d)"
        if log_handler is None:
            log_formatter = logging.Formatter(log_format)
            self.log_handler = logging.StreamHandler()
            self.log_handler.setLevel(log_level)
            self.log_handler.setFormatter(log_formatter)
        else:
            self.log_handler = log_handler
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(self.log_handler)
        self.logger.setLevel(self.log_handler.level)
        self.logger.debug("Logger ready...")
        self.thread_lock = RLock()
        self.access_token = None
        self.access_token_expiry_time = None
        self.available_oi_sort_by = [ "futureswithhighestoi" , "optionswithhighestoi" , "oigainers" , "oilosers",
                                    "oigainerspricegainers" , "oigainerspricelosers" , "oiloserspricegainers" ,
                                    "oiloserspricegainers" ,]
        try:
            self.hist_login()
        except Exception as e:
            self.logger.error(f"Failed to connect REST historical API -> {type(e)} = {e}")

    def get_new_token(self ):
        url_auth = "https://auth.truedata.in/token"
        payload = f"username={self.login_id}&password={self.password}&grant_type=password"
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        token_json = requests.request("POST", url_auth, headers=headers, data=payload).json()
        return token_json

    def hist_login(self):
        token_json = self.get_new_token()
        try:
            if token_json['access_token']:
                self.access_token = token_json['access_token']
                self.access_token_expiry_time = dt.now() + timedelta(seconds=token_json['expires_in'] - 15)  # 15 seconds is random buffer time
                self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Connected successfully to TrueData Anaytical Data Service... {Style.RESET_ALL}")
        except Exception as e:
            self.logger.error(f"Failed to connect -> {token_json['error_description']}{type(e)} = {e}")
            self.access_token = None

    @staticmethod
    def decompress_data(data):
        uncom_length = struct.unpack('<I', data[:4])[0]
        com_length = struct.unpack('<I', data[4:8])[0]
        dc = lz4.block.decompress( data[8:], uncom_length ) if com_length != uncom_length else data[8:]
        return dc.decode()

    @staticmethod
    def clean_data_to_df( data):   
        df = pd.read_csv(StringIO(data) , dtype='unicode' )
        return df

    @access_token_decorator
    def get_option_chain(self, symbol , expiry , greeks = False ):
        try:
            if not isinstance (expiry , date):
                raise TypeError("Expiry must be a type of date") 
            headers = {'Authorization': f'Bearer {self.access_token}'}
            payload = {'symbol' : symbol , "expiry" : expiry.strftime("%d-%m-%Y")}
            end_point = "getoptionchain" if not greeks else "getoptionchainwithgreeks"
            url = f"{self.analytics_url}{end_point}" if not greeks else f"{self.greeks_url}{end_point}"
            with self.thread_lock:
                response = requests.get( url , headers=headers, params=payload)
                if response.status_code == 429:
                    error = base64.b64decode(response.content).decode('utf-8')
                    raise TooManyRequestsError(error)
                elif not response.status_code == 200:
                    raise response.raise_for_status()
                df = self.clean_data_to_df(response.text)
                return df 
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None

    @access_token_decorator
    def get_oi_gainer_losers(self, top , series , sort_by  ):
        try:
            if not sort_by.lower() in self.available_oi_sort_by: 
                raise TypeError("please provide valid sort_by option from the documentation") 
            headers = {'Authorization': f'Bearer {self.access_token}'}
            payload = {'top' : top , "series" : series }
            end_point = f"get{sort_by.lower()}"
            with self.thread_lock:
                response = requests.get(f"{self.analytics_url}{end_point}", headers=headers, params=payload)
                if response.status_code == 429:
                    error = base64.b64decode(response.content).decode('utf-8')
                    raise TooManyRequestsError(error)
                elif not response.status_code == 200:
                    raise response.raise_for_status()
                df = self.clean_data_to_df(response.text)
                return df 
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None
    
    @access_token_decorator
    def get_index_gainer_losers(self, top  , segment , index_name , sort_by  ):
        try:
            if not sort_by.lower() in ["gainers" , "losers"]: 
                raise TypeError("please provide valid sort_by option from the documentation") 
            headers = {'Authorization': f'Bearer {self.access_token}'}
            payload = {'top' : top , "indexname" : index_name , "segment": segment }
            end_point = f"getindex{sort_by.lower()}"
            with self.thread_lock:
                response = requests.get(f"{self.analytics_url}{end_point}", headers=headers, params=payload)
                if response.status_code == 429:
                    error = base64.b64decode(response.content).decode('utf-8')
                    raise TooManyRequestsError(error)
                elif not response.status_code == 200:
                    raise response.raise_for_status()
                df = self.clean_data_to_df(response.text)
                return df 
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None
    
    @access_token_decorator
    def get_industry_gainer_losers(self, top , segment , industry , sort_by  ):
        try:
            if not sort_by.lower() in ["gainers" , "losers"]: 
                raise TypeError("please provide valid sort_by option from the documentation") 
            headers = {'Authorization': f'Bearer {self.access_token}'}
            payload = {'top' : top , "segment": segment , "industry" : industry }
            end_point = f"getindustry{sort_by.lower()}"
            with self.thread_lock:
                response = requests.get(f"{self.analytics_url}{end_point}", headers=headers, params=payload)
                if response.status_code == 429:
                    error = base64.b64decode(response.content).decode('utf-8')
                    raise TooManyRequestsError(error)
                elif not response.status_code == 200:
                    raise response.raise_for_status()
                df = self.clean_data_to_df(response.text)
                return df 
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None
    
    @access_token_decorator
    def get_industry_gainer_losers(self, top , segment , industry , sort_by  ):
        try:
            if not sort_by.lower() in ["gainers" , "losers"]: 
                raise TypeError("please provide valid sort_by option from the documentation") 
            headers = {'Authorization': f'Bearer {self.access_token}'}
            payload = {'top' : top , "segment": segment , "industry" : industry }
            end_point = f"getindustry{sort_by.lower()}"
            with self.thread_lock:
                response = requests.get(f"{self.analytics_url}{end_point}", headers=headers, params=payload)
                if response.status_code == 429:
                    error = base64.b64decode(response.content).decode('utf-8')
                    raise TooManyRequestsError(error)
                elif not response.status_code == 200:
                    raise response.raise_for_status()
                df = self.clean_data_to_df(response.text)
                return df 
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None

    @access_token_decorator
    def get_history_greeks(self, symbol , expiry , strike , series , ltp = False  ):
        try:
            if not isinstance (expiry , date):
                raise TypeError("Expiry must be a type of date") 
            headers = {'Authorization': f'Bearer {self.access_token}'}
            payload = { 'symbol' : symbol , "expiry": expiry.strftime("%d-%m-%Y") ,
                        "strike" : strike, "series" : series }
            end_point = f"getltpwithgreeks" if ltp else "getTickHistorywithGreeks"
            with self.thread_lock:
                response = requests.get(f"{self.greeks_url}{end_point}", headers=headers, params=payload)
                if response.status_code == 429:
                    error = base64.b64decode(response.content).decode('utf-8')
                    raise TooManyRequestsError(error)
                elif not response.status_code == 200:
                    raise response.raise_for_status()
                df = self.clean_data_to_df(response.text)
                return df 
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None

class TooManyRequestsError(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self ):
        return self.msg



tick_fields = [ "timestamp","symbol_id","symbol","ltp","ltq","atp","ttq","day_open","day_high","day_low","prev_day_close",
                     "oi","prev_day_oi","turnover","special_tag","tick_seq","best_bid_price","best_bid_qty","best_ask_price",
                     "best_ask_qty","change","change_perc","oi_change","oi_change_perc" ]

bar_fields = ["symbol" ,"symbol_id",  "timestamp", "bar_open", "bar_high", "bar_low", "bar_close", "bar_volume", "oi", "ttq" ]

bid_ask_fields =    [ "timestamp", "symbol_id", "symbol", "bid", "ask", "total_bid", "total_ask" ]

greek_fields = [ "timestamp", "symbol_id", "symbol", "iv", "delta", "theta", "gamma", "vega", "rho" ] 

full_feed_fields = [ "timestamp", "symbol_id", "symbol", "ltp", "ltq", "atp", "ttq", "day_open", "day_high", 
                    "day_low" , "prev_day_close", "oi", "prev_day_oi", "turnover", "tick_seq", "best_bid_price",
                    "best_bid_qty", "best_ask_price", "best_ask_qty" ]

min_fields = [ "symbol", "symbol_id", "day_open", "day_high", "day_low", "prev_day_close", "prev_day_oi", "oi",
        "ttq", "timestamp", "open", "high", "low", "close", "volume", "change","change_perc","oi_change","oi_change_perc" ]

chain_columns = ['symbols' , 'strike' , 'type' , 'ltp' , 'ltt' , 'ltq' , 'volume' , 'price_change' , 'price_change_perc', 'oi' , 
                'prev_oi' , 'oi_change' , 'oi_change_perc' , 'bid' , 'bid_qty' , 'ask' , 'ask_qty']

chain_live_update_column = ['ltp' , 'ltt'  , 'ltq' , 'volume' ,'price_change' , 'price_change_perc' , 'oi' , 'prev_oi',
                'oi_change' , 'oi_change_perc','bid' , 'bid_qty' , 'ask' , 'ask_qty' ]

chain_min_update_column = chain_live_update_column[:-4]

chain_fields_from_live_dict = [ "ltp", "timestamp", "ltq", "ttq", "change", "change_perc", "oi", "prev_day_oi",
                                "oi_change", "oi_change_perc", "best_bid_price", "best_bid_qty","best_ask_price", "best_ask_qty" ]

chain_fields_from_min_dict = [ "close", "timestamp", "volume", "ttq", "change", "change_perc", "oi", "prev_day_oi",
                                "oi_change", "oi_change_perc",  ]

chain_greek_fields =  ["iv", "delta", "theta", "gamma", "vega", "rho" ]